/*-
 * Copyright (C)2003 @BABOLO http://www.babolo.ru/
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 *  OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ifndef lint
static const char copyright[] = "\
@(#)Copyright (C)2003 @BABOLO http://www.babolo.ru/\n\
@(#)All rights reserved.\n";
static const char rcsid[] = "$Id: read.c,v 1.15 2011/11/15 00:24:42 babolo Exp $";
#endif /* not lint */

#include <machine/param.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <sys/mman.h>
#include <sys/stat.h>
#include <sys/uio.h>
#include <limits.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <stdio.h>
#include <err.h>
#include <babolo/BLINflag.h>
#include "mife.h"
#include "mife_int.h"

#define COND_NULL_BUF  0x0001  /*                            */
#define COND_BIG_GAP   0x0002  /*      */
#define COND_NEED_REM  0x0004  /*   ,  remalloc   */
#define COND_NEED_MOV  0x0008  /*        */
#define COND_NEED_RD   0x0010  /*                  */

#define C_fr 0x01 /* free buffer     */
#define C_zr 0x02 /* empty buffer    */
#define C_al 0x04 /* allocate buffer */
#define C_mv 0x08 /* shift buffer    */
#define C_ra 0x10 /* realloc buffer  */
#define C_in 0x20 /* read to buffer  */

static const u_char nuldecod[] =
{ 0      /* 00 */
, C_al   /* 01 */
, C_in   /* 02 */
, C_al   /* 03 */
};
static const u_char pipdecod[] =
{ 0                         /* 00 */
,        C_zr | C_al        /* 01 */
,        C_zr               /* 02 */
,        C_zr | C_al        /* 03 */
,        C_ra               /* 04 */
,        C_zr | C_al        /* 05 */
,        C_fr | C_zr | C_al /* 06 */
,        C_zr | C_al        /* 07 */
,        C_mv               /* 08 */
,        C_zr | C_al        /* 09 */
,        C_zr               /* 0A */
,        C_zr | C_al        /* 0B */
,        C_mv | C_ra        /* 0C */
,        C_zr | C_al        /* 0D */
,        C_fr | C_zr | C_al /* 0E */
,        C_zr | C_al        /* 0F */
, C_in                      /* 10 */
, C_in | C_zr | C_al        /* 11 */
, C_in | C_zr               /* 12 */
, C_in | C_zr | C_al        /* 13 */
, C_in | C_ra               /* 14 */
, C_in | C_zr | C_al        /* 15 */
, C_in | C_fr | C_zr | C_al /* 16 */
, C_in | C_zr | C_al        /* 17 */
, C_in | C_mv               /* 18 */
, C_in | C_zr | C_al        /* 19 */
, C_in | C_zr               /* 1A */
, C_in | C_zr | C_al        /* 1B */
, C_in | C_mv | C_ra        /* 1C */
, C_in | C_zr | C_al        /* 1D */
, C_in | C_fr | C_zr | C_al /* 1E */
, C_in | C_zr | C_al        /* 1F */
};

ssize_t
mife_read(mife_descriptor *mife, ssize_t ordlen, off_t offset) {
    ifBLIN_FLAG = mife ? mife->flag : 0;
    struct stat sb;
    struct rlimit rlp;
    int retrylim, plb;
    u_int32_t cond;
    u_int32_t control;
    off_t delta;
    ssize_t needlen = ordlen;

    ifBLIN_QX2( "+ %s needlen=%08X offset=%016llX"
              , mife ? "OK" : "NULL"
              , (u_int)needlen
              , (long long)offset
              );
    if  (!mife) {
        errno = EINVAL;
        // XXXX warn("mife=NULL");
        return(-1);
    }
    ifBLIN_QX3( "mife->flag=%08X fd=%d mindelta=%08X\n"
                "mife->actlen=%08X  buflen=%08X offset=%08X"
              , mife->flag, mife->fd, (u_int)mife->mindelta
              , (u_int)mife->actlen, (u_int)mife->buflen, (u_int)mife->offset
              );
    plb = (mife->flag & MIFE_FULL) ? 1 : 0;
    if  (needlen > SSIZE_MAX) { /* XXXX NoOp */
        errno = EDOM;
        ifBLIN_QW0("#03");
        return(-1);
    }
    if  (needlen < 0) {
        errno = EINVAL;
        ifBLIN_QW0("#04");
        return(-1);
    }
    delta = offset - mife->offset;
    if  (delta < (off_t)0) {
        errno = ESPIPE;
        ifBLIN_QW0("#05 pipe");
        return(-1);
    }
    if  ((mife->flag & MIFE_EOFL) || ((needlen != 0) && (mife->actlen >= (needlen + plb + delta)))) {
        ifBLIN_QX2("- #06 null %08X", (unsigned)(mife->actlen - delta));
        return(mife->actlen - delta);
    }
    ifBLIN_QX3("needlen=%08X plb=%d", (u_int)needlen, plb);
    if  (mife->flag & MIFE_INMM) {
        ifBLIN_QX3(" -> null");
        cond = 0;
        if  (!mife->buffer) cond |= COND_NULL_BUF;
        if  (mife->actlen < delta) cond |= COND_BIG_GAP;
        control = nuldecod[cond];
        if  (control & C_in) {
            errno = ESPIPE;
            ifBLIN_QW0("#11");
            return(-1);
        }
        if  (control & C_al) {
            errno = EINVAL;
            ifBLIN_QX0("#12 pipe: internal software error");
            return(-1);
        }
        ifBLIN_QX2("- #12 null %08X", (unsigned)(mife->actlen - delta));
        return(mife->actlen - delta);
    }
    if  (!mife->st_mode) {
        if  (fstat(mife->fd, &sb)) {
            ifBLIN_QW0("#19");
            return(-1);
        }
        mife->st_mode = sb.st_mode;
        mife->st_size = sb.st_size;
    }
    ifBLIN_QX3("st_mode=%011o", mife->st_mode); /* XXXX delete \n */
    switch (mife->st_mode & S_IFMT) {
    case S_IFREG: {
        if  (needlen == 0 && !(mife->flag & MIFE_SLID)) needlen = mife->st_size;
        ifBLIN_QX3(" -> file=%u", (unsigned)mife->st_size);
    ;   if  (mife->flag & MIFE_INMM) {
            errno = EINVAL;
            ifBLIN_QX0("#20 pipe: internal software error");
            return(-1);
        }
        if  (!(mife->flag & (MIFE_ALLC | MIFE_PIPE))) {
            /* XXXX Not yet */
            ssize_t tmplen;
            off_t tmpsz;

            /* remmap */
            if  (mife->buffer && munmap(mife->buffer, mife->buflen) < 0) {
                ifBLIN_QW2("- #21: file");
                return(-1);
            }
            /* Biggest tail for this file and offset */
            tmpsz = mife->st_size - offset + plb;
            if  (tmpsz <= 0) {
                tmplen = (tmpsz < -SSIZE_MAX) ? -SSIZE_MAX : tmpsz;
                mife->buffer = NULL;
                mife->actlen = 0;
                mife->buflen = 0;
                mife->offset = offset;
                ifBLIN_QX2("- #22 file %d", (int)tmplen);
                return(tmplen);
            }
            /* XXXX */ rlp.rlim_cur = 0; /* XXXX just err out */
            tmplen = ((rlp.rlim_cur == RLIM_INFINITY) ? DEFMEM : rlp.rlim_cur >> 8) + mife->mindelta;
            if  (mife->flag & MIFE_BUFX) tmplen = mife->mindelta;
            if  (tmpsz < tmplen) tmplen = tmpsz;
            mife->buffer = mmap(NULL, tmplen, PROT_READ | PROT_WRITE, MAP_PRIVATE, mife->fd, offset);
            if  (mife->buffer == MAP_FAILED) {
                ifBLIN_QX2("-#23: file: %s", strerror(errno));
                return(-1);
            }
            mife->actlen = needlen;
            mife->buflen = tmplen;
            mife->offset = offset;
            if  (mife->flag & MIFE_FULL) ((u_char *)(mife->buffer))[mife->actlen] = 0;
            if  (mife->flag & MIFE_FULL && mife->flag & BLIN_VER4)
                fprintf( stderr, "===================\n%s~~~~~~~~~~~~~~~~~~~\n"
                       , (u_char *)(mife->buffer)
                       );
            ifBLIN_QX2("- #24 file %08X", (u_int)tmplen);
            return(tmplen);
    }   }
        /* FALLTHRU */
    case S_IFIFO:
    case S_IFCHR:
    case S_IFBLK:
    case S_IFSOCK: {
        ssize_t nearlen;

    ;   nearlen = ((needlen > mife->mindelta) ? needlen : mife->mindelta);
    ;   ifBLIN_QX3(" -> pipe, nearlen=%08X, eomax=%d", (u_int)nearlen, (int)mife->eomax);
    ;   if  (mife->flag & (MIFE_INMM | MIFE_MMAP)) {
            errno = EINVAL;
            ifBLIN_QX0("#31 pipe: internal software error");
            return(-1);
        }
    ;   do {
        ;   delta = offset - mife->offset;
        ;   if  (delta < (off_t)0) {
        ;       errno = ESPIPE;
        ;       ifBLIN_QW0("#32 pipe");
        ;       return(-1);
        ;   }
        ;   cond = 0;
        ;   control = 0;
        ;   if  (!mife->buffer) cond |= COND_NULL_BUF;
        ;   if  (needlen == 0 && delta <= mife->actlen)
                while (mife->actlen - delta >= nearlen - mife->mindelta) nearlen += mife->mindelta;
        ;   if  (mife->buflen - delta - plb < nearlen) cond |= COND_NEED_REM;
        ;   if  (mife->buflen - mife->actlen < plb) cond |= COND_NEED_REM;
        ;   if  (mife->actlen - delta < nearlen) {
        ;       cond |= COND_NEED_RD;
        ;       if  (mife->buflen - mife->actlen - plb < nearlen) cond |= COND_NEED_REM;
        ;       if  (mife->buflen < delta + MAXPACKT + plb) cond |= COND_NEED_MOV;
        ;       if  (mife->buflen < mife->actlen + MAXPACKT + plb) cond |= COND_NEED_MOV;
        ;       if  (mife->actlen < delta) cond |= COND_BIG_GAP;
        ;       if  (delta > MAXPACKT) cond |= COND_NEED_MOV;
        ;       if  (mife->flag & MIFE_BUFX && (delta > PAGE_SIZE)) cond |= COND_NEED_MOV;
        ;   }
        ;   control = pipdecod[cond];
        ;   ifBLIN_QX3( "mife->actlen=%08X  buflen=%08X offset=%08X\n"
                        "delta=%08X nearlen=%08X cond=%02X control=%02X"
                      , (u_int)mife->actlen, (u_int)mife->buflen, (u_int)mife->offset
                      , (u_int)delta, (u_int)nearlen, cond, control
                      );
        ;   if  (control & C_fr) {
        ;       ifBLIN_QX4("free");
        ;       free(mife->buffer);
        ;       mife->buffer = NULL;
        ;       eolstop(mife);
        ;       mife->buflen = 0;
        ;       mife->flag &= ~MIFE_ALLC;
        ;   }
        ;   if  (control & C_zr) {
        ;       ifBLIN_QX4("zero");
        ;       eolstop(mife);
        ;       mife->offset += mife->actlen;
        ;       mife->actlen = 0;
        ;   }
        ;   if  (control & C_al) {
        ;       mife->buflen = (((unsigned)nearlen - 1 + plb) | (MAXPACKT - 1)) + 1;
        ;       ifBLIN_QX4("malloc(%08X)", (u_int)mife->buflen);
        ;       eolstop(mife);
        ;       mife->actlen = 0;
        ;       mife->flag |= MIFE_ALLC;
        ;       if  (!(mife->buffer = malloc(mife->buflen))) {
        ;           errno = ENOMEM;
        ;           ifBLIN_QX0("#33 pipe");
        ;           return(-1);
        ;   }   }
        ;   if  (control & C_mv) {
        ;       ifBLIN_QX4("need move");
        ;       bcopy(mife->buffer + delta, mife->buffer, mife->actlen - delta);
        ;       eoldelta(mife, -delta);
        ;       mife->offset += delta;
        ;       mife->actlen -= delta;
        ;       delta = 0;
        ;   }
        ;   if  (control & C_ra) {
        ;       mife->buflen = (((unsigned)nearlen - 1 + plb) | (MAXPACKT - 1)) + 1 + mife->actlen;
        ;       ifBLIN_QX4("reallocf(%08X)", (u_int)mife->buflen);
        ;       if  (!(mife->buffer = reallocf(mife->buffer, mife->buflen))) {
        ;           errno = EFBIG;
        ;           ifBLIN_QX0("#34 pipe");
        ;           return(-1);
        ;   }   }
        ;   eolstart(mife, delta);
        ;   if  (mife_eoldetect(mife) >= 0) {
        ;       ifBLIN_QX3("p eolp=%d, eolst=%d", (int)mife->eolp, (int)mife->eolst);
        ;       control = 0;
        ;   }
        ;   if  (control & C_in) {
                ssize_t tmplen;
        ;       ifBLIN_QX4( "input: delta=%08X buflen=%08X needlen=%08X nearlen=%08X"
                          , (u_int)delta, (u_int)mife->buflen, (u_int)needlen, (u_int)nearlen
                          );
        ;       retrylim = getMIFE_RTRY(mife->flag);
        ;       do { /* Read something only onse */
        ;           tmplen = read( mife->fd
                                 , mife->buffer + mife->actlen
                                 , (needlen > MAXPACKT) ? needlen : MAXPACKT
                                 );
        ;           ifBLIN_QX5( "retrylim=%d tmplen=%08X nearlen=%08X"
                              , retrylim, (u_int)tmplen, (u_int)nearlen
                              );
        ;           if  (tmplen < 0) {
        ;               if  (errno == EAGAIN || errno == EDEADLK) {
        ;                   ifBLIN_QW1("#35 pipe");
        ;                   tmplen = 0;
        ;               } else {
        ;                   ifBLIN_QW0("#36 pipe");
        ;                   return(-1);
        ;               }
        ;           } else if (tmplen == 0) {
        ;               if  (retrylim-- <= 0 || S_ISREG(mife->st_mode)) break;
        ;               usleep(1 << ((getMIFE_RTRY(mife->flag)) - retrylim));
        ;           } else {
        ;               mife->actlen += tmplen;
        ;               eolstart(mife, delta);
        ;               if  (mife_eoldetect(mife) >= 0) {
        ;                   ifBLIN_QX4(" eolp=%d, eolst=%d", (int)mife->eolp, (int)mife->eolst);
        ;                   control = 0;
        ;                   break;
        ;           }   }
        ;       } while (tmplen == 0);
        ;       if  (tmplen == 0) {
        ;           mife->flag |= MIFE_EOFL;
        ;           control = 0;
        ;   }   }
        } while (control != 0);
    ;   {   ssize_t tmplen;
        ;   tmplen = mife->buflen - mife->actlen;
        ;   delta = offset - mife->offset;
        ;   ifBLIN_QX3( "mife->actlen=%08X  buflen=%08X offset=%08X delta=%08X"
                      , (u_int)mife->actlen, (u_int)mife->buflen, (u_int)mife->offset, (u_int)delta
                      );
        ;   if  (tmplen < plb) {
        ;       errno = EINVAL;
        ;       ifBLIN_QX0("#37 pipe: internal software error");
        ;       return(-1);
        ;   }
        ;   if  (delta > mife->actlen) {
        ;       errno = ESPIPE;
        ;       ifBLIN_QW0("#38 pipe");
        ;       return(-1);
        ;   }
        ;   if  (tmplen > 0) bzero(mife->buffer + mife->actlen, (tmplen > 8) ? 8 : tmplen);
        ;   ifBLIN_QX2( "- #39 pipe %08X(%d)", (unsigned)(mife->actlen - delta), (int)tmplen);
        ;   return((ssize_t)mife->actlen - delta);
    }   }
    case S_IFDIR:
    case S_IFLNK:
    case S_IFWHT:
    default:
        errno = EINVAL;
        ifBLIN_QW0(" -> unknown");
        ifBLIN_QX0("- #40");
        return(-1);
}   }
